import sys
from IPython.terminal.embed import InteractiveShellEmbed, load_default_config, InteractiveShell

# NOTE: I had to copy the code so I could add stack_depth
def embed(*, header="", compile_flags=None, stack_depth=0, **kwargs):
    config = kwargs.get('config')
    if config is None:
        config = load_default_config()
        config.InteractiveShellEmbed = config.TerminalInteractiveShell
        kwargs['config'] = config
    using = kwargs.get('using', 'sync')
    if using :
        kwargs['config'].update({'TerminalInteractiveShell':{'loop_runner':using, 'colors':'NoColor', 'autoawait': using!='sync'}})
    #save ps1/ps2 if defined
    ps1 = None
    ps2 = None
    try:
        ps1 = sys.ps1
        ps2 = sys.ps2
    except AttributeError:
        pass
    #save previous instance
    saved_shell_instance = InteractiveShell._instance
    if saved_shell_instance is not None:
        cls = type(saved_shell_instance)
        cls.clear_instance()
    frame = sys._getframe(1)
    shell = InteractiveShellEmbed.instance(_init_location_id='%s:%s' % (
        frame.f_code.co_filename, frame.f_lineno), **kwargs)
    shell(header=header, stack_depth=2+stack_depth, compile_flags=compile_flags,
        _call_location_id='%s:%s' % (frame.f_code.co_filename, frame.f_lineno))
    InteractiveShellEmbed.clear_instance()
    #restore previous instance
    if saved_shell_instance is not None:
        cls = type(saved_shell_instance)
        cls.clear_instance()
        for subclass in cls._walk_mro():
            subclass._instance = saved_shell_instance
    if ps1 is not None:
        sys.ps1 = ps1
        sys.ps2 = ps2


import time
import mqdm as M
from mqdm import mqdm, pool, print


# ---------------------------------------------------------------------------- #
#                                   Examples                                   #
# ---------------------------------------------------------------------------- #


def example_prompt():
    for i in mqdm(range(10), desc='example', transient=False):
        time.sleep(0.1)
        if i == 5:
            M.inp("Do you want to continue?")


def example_group():
    t0 = time.time()
    for i in range(5):
        for j in mqdm(range(100), desc=f'blah {i}'):
            time.sleep(0.005)
        print("loop")
    print(f"done in {time.time() - t0:.2f} seconds")

    t0 = time.time()
    with M.group():
        for i in range(5):
            for j in mqdm(range(100), desc=f'blah {i}'):
                time.sleep(0.005)
            print("loop")
    print(f"done in {time.time() - t0:.2f} seconds")

    t0 = time.time()
    for i in range(5):
        for j in mqdm(range(100), desc=f'blah {i}'):
            time.sleep(0.005)
        print("loop")
    print(f"done in {time.time() - t0:.2f} seconds")



def example_bar(n=8, sleep=1, transient=False, error=False, indet=False, embed=False, bp=False):
    t0 = time.time()
    xs = range(n)
    if indet:
        xs = (x for x in xs)
    for i in M.mqdm(xs, desc='example', transient=transient):
        M.set_description(f'example {i}')
        for j in M.mqdm(range(10), desc=f'blah {i}', transient=transient):
            time.sleep(0.04 * sleep)
            if j == 5 and not i % 2:
                # print("blah", i, j)
                if error: 1/0
                if embed: M.embed()
                if bp: M.bp()
    print(f"done in {time.time() - t0:.2f} seconds")


def example_fn(n, error=False, sleep=1):
    import time
    import random
    for i in mqdm(range(n + 1), desc=f'example {n}'):
        t = sleep * random.random()*2 / (i+1)
        time.sleep(t)
        print(i, "slept for", t)
        # mqdm_.set_description("sleeping for %.2f" % t)
        if error: 1/0
    print("Done", n)

def example_pool(n=5, transient=False, n_workers=5, **kw):
    import time
    t0 = time.time()
    pool(
        example_fn,
        # example_bar, 
        range(n), 
        '[bold blue]Very important work',
        bar_kw={'transient': transient},
        n_workers=n_workers,
        **kw)
    print("done in", time.time() - t0, "seconds", 123)

def example_cf_pool(n=5, transient=False, n_workers=5, **kw):
    from concurrent.futures import ProcessPoolExecutor, as_completed
    import time
    t0 = time.time()

    it = range(n)
    with ProcessPoolExecutor(max_workers=n_workers, initializer=M.Initializer("process")) as executor, \
         mqdm(desc='[bold blue]Very important work', total=len(it), transient=transient) as pbar:
        futures = {executor.submit(example_fn, i, **kw): i for i in it}
        for future in as_completed(futures):
            result = future.result()
            pbar.update(1)
    print("done in", time.time() - t0, "seconds", 123)

def example_mp_pool(n=5, transient=False, n_workers=5, **kw):
    from multiprocessing import Pool
    import time
    t0 = time.time()

    it = range(n)
    with Pool(processes=n_workers, initializer=M.Initializer("process")) as executor, \
         mqdm(desc='[bold blue]Very important work', total=len(it), transient=transient) as pbar:
        for _ in executor.imap_unordered(example_fn, it):
            pbar.update(1)
    print("done in", time.time() - t0, "seconds", 123)

def example_cf_thread_pool(n=5, transient=False, n_workers=5, **kw):
    from concurrent.futures import ThreadPoolExecutor, as_completed
    import time
    t0 = time.time()

    it = range(n)
    with ThreadPoolExecutor(max_workers=n_workers) as executor, \
         mqdm(desc='[bold blue]Very important work', total=len(it), transient=transient) as pbar:
        futures = {executor.submit(example_fn, i, **kw): i for i in it}
        for future in as_completed(futures):
            result = future.result()
            pbar.update(1)
    print("done in", time.time() - t0, "seconds", 123)

def example_thread_pool(n=5, transient=False, n_workers=5, **kw):
    from threading import Thread
    import time
    t0 = time.time()

    it = range(n)
    with mqdm(desc='[bold blue]Very important work', total=len(it), transient=transient) as pbar:
        threads = []
        def worker(i):
            example_fn(i, **kw)
            pbar.update(1)
        for i in it:
            t = Thread(target=worker, args=(i,))
            t.start()
            threads.append(t)
        for t in threads:
            t.join()
    print("done in", time.time() - t0, "seconds", 123)


import logging
from mqdm import install_logging

install_logging(level=logging.INFO)
logger = logging.getLogger(__name__)


def log_fn(n, sleep=1):
    import random
    for i in mqdm(range(n + 1), desc=f'example {n}'):
        t = sleep * random.random()*2 / (i+1)
        time.sleep(t)
        logger.info(f"Task {n}, iteration {i}, slept for {t:.2f} seconds")
    logger.info(f"Done {n}")

def example_logging(n=5, transient=False, n_workers=5, **kw):
    pool(
        log_fn,
        range(n), 
        '[bold blue]Very important work with logging',
        bar_kw={'transient': transient},
        n_workers=n_workers,
        **kw)

# @M.profile
def speed_fn(t, N=1000000000):
    # print("Starting")
    import time
    t0 = time.time()
    for i in mqdm(range(N), desc=M.utils.process_name()):
        if time.time() - t0 > t:
            # print("break", i)
            break

# @M.utils.profile
def example_speed(t=10, n_workers=1, x=1, **kw):
    import time
    t0 = time.time()
    pool(
        speed_fn,
        # example_bar, 
        [t]*n_workers*x, 
        '[bold blue]Very important work',
        bar_kw={'transient': False},
        n_workers=n_workers,
        squeeze_=False,
        **kw)
    print("done in", time.time() - t0, "seconds", 123)
    time.sleep(2)


def tqdm_speed_fn(t, N=1000000000):
    import tqdm
    import time
    t0 = time.time()
    for i in tqdm.tqdm(range(N)):
        if time.time() - t0 > t:
            print("break")
            break

def example_tqdm_speed(t=10, transient=False, **kw):
    import time
    t0 = time.time()
    pool(
        tqdm_speed_fn,
        # example_bar, 
        [t]*1, 
        '[bold blue]Very important work',
        bar_kw={'transient': transient},
        # transient=True,
        n_workers=1,
        **kw)
    print("done in", time.time() - t0, "seconds", 123)


def example_messy(n=3, transient=False, n_workers=5, **kw):
    import time
    t0 = time.time()
    example_bar(2, transient=False)
    pbar = mqdm(desc='example leftover', leave=True)
    # pbar1 = mqdm(desc='other', total=10)
    # with pbar:
    pbar.update()
    # time.sleep(1)
    pbar.update()
    # time.sleep(1)
    # pbar.__enter__()
    pbar.update()

    pool(
        example_fn, 
        range(n), 
        '[bold blue]Very important work',
        bar_kw={'transient': True},
        n_workers=n_workers,
        **kw)
    pbar.update()
    pbar.update()
    time.sleep(1)
    example_bar(3, sleep=1, transient=True)
    time.sleep(1)
    pbar.update(total=3)
    # pbar.__exit__(None, None, None)
    # time.sleep(1)
    # mqdm_.pbar.remove_task(0)
    # mqdm_.pbar.stop()
    # time.sleep(1)
    # example(2, transient=False)
    print("done in", time.time() - t0, "seconds", 123)
    pbar.close()
    # M.pbar.stop()
    # time.sleep(1)
    print("done in", time.time() - t0, "seconds", 123)
    print("done in", time.time() - t0, "seconds", 123)
    # print("done in", time.time() - t0, "seconds", 123)
    # print("done in", time.time() - t0, "seconds", 123)
    # print("done in", time.time() - t0, "seconds", 123)
    # print("done in", time.time() - t0, "seconds", 123)
    # 1/0

    # example_bar(2, transient=False)
    # M.pbar.stop()


def main():
    example_prompt()
    example_group()
    example_bar(n=8, sleep=0.1, transient=False)
    example_bar(n=8, sleep=0.1, transient=True)
    try:
        example_bar(n=8, sleep=0.1, transient=False, error=True)
    except ZeroDivisionError:
        pass
    example_pool(n=3, transient=False, n_workers=2)
    example_pool(n=3, transient=True, n_workers=2)
    example_speed(t=5, n_workers=2, x=2)
    example_tqdm_speed(t=5, transient=False)
    example_tqdm_speed(t=5, transient=True)
    example_messy(n=3, transient=False, n_workers=2)


if __name__ == '__main__':
    all = main
    import fire
    fire.Fire()


import logging
import warnings
import mqdm as M


class MQDMHandler(logging.Handler):
    """A logging handler that routes records through mqdm.print so output
    is rendered above progress bars and works across processes.

    Note: formats to plain strings for safe cross-process transmission.
    """
    LEVEL_STYLE = {
        logging.DEBUG: "dim",
        logging.INFO: "",
        logging.WARNING: "yellow",
        logging.ERROR: "red",
        logging.CRITICAL: "bold red",
    }

    def __init__(self, *, markup: bool = True, level: int = logging.NOTSET):
        super().__init__(level)
        self.markup = markup

    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            if self.markup:
                style = self.LEVEL_STYLE.get(record.levelno, "")
                if style:
                    msg = f"[{style}]{msg}[/{style}]"
            M.print(msg)
        except Exception:
            self.handleError(record)


def _make_default_formatter() -> logging.Formatter:
    return logging.Formatter(
        fmt="%(asctime)s | %(levelname)s | %(name)s: %(message)s",
        datefmt="%H:%M:%S",
    )


def install(
    logger: logging.Logger = None,
    *,
    level: int = None,
    replace_root: bool = False,
    capture_warnings: bool = True,
    markup: bool = True,
    formatter: logging.Formatter | None = None,
) -> None:
    """Install an mqdm-aware logging handler on the root logger.

    - replace_root: if True, clears existing root handlers before adding ours.
    - capture_warnings: if True, directs Python warnings to the logging system.
    - formatter: optional logging.Formatter; a sensible default is used otherwise.

    Also stores minimal configuration so worker processes can mirror logging.
    """
    if logger is None:
        logger = logging.getLogger()
    if replace_root:
        logger.handlers.clear()

    # Avoid duplicate handlers; reuse if already installed
    existing = next((h for h in logger.handlers if isinstance(h, MQDMHandler)), None)
    handler = existing or MQDMHandler(markup=markup)
    handler.setFormatter(formatter or getattr(handler, 'formatter', None) or _make_default_formatter())
    if not existing:
        logger.addHandler(handler)
    if level is not None:
        logger.setLevel(level)

    if capture_warnings:
        warnings.simplefilter("default")
        logging.captureWarnings(True)

    # Save minimal config for worker processes to mirror
    M._logging_config = {
        "level": level,
        "markup": markup,
        "capture_warnings": capture_warnings,
        "formatter_fmt": (formatter._fmt if formatter else None),
        "formatter_datefmt": (formatter.datefmt if formatter else None),
        "replace_root": False,  # never replace in workers by default
    }


def uninstall() -> None:
    """Remove MQDMHandler from the root logger and stop warning capture."""
    logger = logging.getLogger()
    to_remove = [h for h in logger.handlers if isinstance(h, MQDMHandler)]
    for h in to_remove:
        logger.removeHandler(h)
    logging.captureWarnings(False)


def _install_from_config(cfg: dict | None) -> None:
    """Internal: install handler in workers using stored config from mqdm._logging_config."""
    if not cfg:
        return
    formatter = None
    if cfg.get("formatter_fmt"):
        formatter = logging.Formatter(cfg["formatter_fmt"], cfg.get("formatter_datefmt"))
    install(
        level=cfg.get("level"),
        replace_root=cfg.get("replace_root", False),
        capture_warnings=cfg.get("capture_warnings", True),
        markup=cfg.get("markup", True),
        formatter=formatter,
    )



import os
from typing import Literal
from concurrent.futures._base import FINISHED, RUNNING
from concurrent.futures import Future, Executor, ThreadPoolExecutor, ProcessPoolExecutor
from concurrent.futures._base import LOGGER
from concurrent.futures.process import _RemoteTraceback  # used in bar.py
import mqdm as M
from ._logging import _install_from_config

# ---------------------------------------------------------------------------- #
#                                 Monkey Patch                                 #
# At some point, _max_tasks_per_child and exit_pid were added to ProcessPoolExecutor.
# At some point we can remove this monkey patch if we drop support for older Python versions.
# ---------------------------------------------------------------------------- #

# from concurrent.futures.process import _sendback_result
from concurrent.futures.process import _ResultItem, _ExceptionWithTraceback
class _ResultItem(_ResultItem):
    def __init__(self, work_id, exception=None, result=None, exit_pid=None):
        self.work_id = work_id
        self.exception = exception
        self.result = result
        self.exit_pid = exit_pid


def _sendback_result(result_queue, work_id, result=None, exception=None, exit_pid=None):
    try:
        result_queue.put(_ResultItem(work_id, result=result, exception=exception, exit_pid=exit_pid))
    except BaseException as e:
        exc = _ExceptionWithTraceback(e, e.__traceback__)
        result_queue.put(_ResultItem(work_id, exception=exc, exit_pid=exit_pid))


def _process_worker(call_queue, result_queue, initializer, initargs, max_tasks=None):
    if initializer is not None:
        try:
            initializer(*initargs)
        except BaseException:
            LOGGER.critical('Exception in initializer:', exc_info=True)
            return
    num_tasks = 0
    exit_pid = None
    while True:
        call_item = call_queue.get(block=True)
        if call_item is None:
            # Wake up queue management thread
            result_queue.put(os.getpid())
            return

        if max_tasks is not None:
            num_tasks += 1
            if num_tasks >= max_tasks:
                exit_pid = os.getpid()

        try:
            r = call_item.fn(*call_item.args, **call_item.kwargs)
        except KeyboardInterrupt as e:  # +++++ Added: ensure KeyboardInterrupt stops the worker
            exc = _ExceptionWithTraceback(e, e.__traceback__)
            _sendback_result(result_queue, call_item.work_id, exception=exc, exit_pid=os.getpid())
            return
        except BaseException as e:
            exc = _ExceptionWithTraceback(e, e.__traceback__)
            _sendback_result(result_queue, call_item.work_id, exception=exc, exit_pid=exit_pid)
        else:
            _sendback_result(result_queue, call_item.work_id, result=r, exit_pid=exit_pid)
            del r

        del call_item

        if exit_pid is not None:
            return


class ProcessPoolExecutor(ProcessPoolExecutor):
    _max_tasks_per_child = None  # specific python version compatibility (?)

    def _spawn_process(self):
        p = self._mp_context.Process(
            target=_process_worker,
            args=(
                self._call_queue,
                self._result_queue,
                self._initializer,
                self._initargs,
                self._max_tasks_per_child,
            ),
        )
        p.start()
        self._processes[p.pid] = p

# ---------------------------------------------------------------------------- #
#                                  Sequential                                  #
# ---------------------------------------------------------------------------- #


class SequentialFuture(Future):
    def __init__(self, fn, *args, **kwargs):
        super().__init__()
        self._fn = fn
        self._args = args
        self._kwargs = kwargs
        self._evaluated = False
        self._processes = {}
        with self._condition:  # so as_completed will return it
            self._state = FINISHED

    def _evaluate(self):
        _rich_traceback_omit = True
        if not self._evaluated:
            with self._condition:
                self._state = RUNNING
            try:
                self.set_result(self._fn(*self._args, **self._kwargs))
            except Exception as exc:
                self.set_exception(exc)
            self._evaluated = True

    def result(self, timeout=None):
        _rich_traceback_omit = True
        self._evaluate()
        return super().result(timeout)
    
    def exception(self, timeout=None):
        _rich_traceback_omit = True
        self._evaluate()
        return super().exception(timeout)


class SequentialExecutor(Executor):
    def __init__(self, max_workers=None, initializer=None, initargs=()):
        super().__init__()
        self._initializer = initializer
        self._initargs = initargs

    def __enter__(self):
        if self._initializer is not None:
            self._initializer(*self._initargs)
        return super().__enter__()

    def submit(self, fn, *args, **kwargs):
        return SequentialFuture(fn, *args, **kwargs)



# ---------------------------------------------------------------------------- #
#                         Concurrent Futures Executors                         #
# ---------------------------------------------------------------------------- #


T_POOL_MODE = Literal['process', 'thread', 'sequential', None]
POOL_EXECUTORS = {
    'thread': ThreadPoolExecutor,
    'process': ProcessPoolExecutor,
    'sequential': SequentialExecutor,
    None: SequentialExecutor,
}


def executor(pool_mode: T_POOL_MODE='process', bar_kw: dict=None, **kw) -> Executor:
    """Return the appropriate executor for the pool mode of the progress bar."""
    return POOL_EXECUTORS[pool_mode](initializer=Initializer(pool_mode, bar_kw), **kw)

import threading
_thread_local_data = threading.local()
def _get_local(key, default=None):
    """Get a thread-local variable."""
    return getattr(_thread_local_data, key, default)


class Initializer:
    def __init__(self, pool_mode: T_POOL_MODE='process', defaults: dict=None):
        self.pool_mode = pool_mode
        self.pbar = M._get_pbar(pool_mode=pool_mode)
        self.pause_event = M._pause_event
        self.shutdown_event = M._shutdown_event
        self.logging_config = M._logging_config
        self.defaults = defaults if defaults is not None else {}
        self.pause_event.set()
        self.shutdown_event.set()

    def __call__(self):
        """Initialize the progress bar for the worker thread/process."""
        M.pbar = self.pbar
        M._pause_event = self.pause_event
        M._shutdown_event = self.shutdown_event
        _thread_local_data.defaults = self.defaults
        # Mirror logging configuration in workers if present
        try:
            _install_from_config(self.logging_config)
        except Exception:
            pass



import sys
import time
import functools
from types import TracebackType
import rich
from rich.prompt import Prompt
from rich.console import Text

import mqdm as M

# ---------------------------------------------------------------------------- #
#                                  Debug Tools                                 #
# ---------------------------------------------------------------------------- #


def embed(*a, prompt='ipython?> ', exit_prompt=True):
    """Embed an IPython shell in the current environment. This will make sure the progress bars don't interfere.
    
    This function is useful for debugging and interactive exploration.

    Does not work in subprocesses for obvious reasons.

    .. code-block:: python

        import mqdm

        for i in mqdm_.mqdm(range(10)):
            if i == 5:
                mqdm_.embed()
    """
    with M.pause():
        from ._embed import embed
        if not prompt or _Prompt.ask(Text(f'{prompt}', style="dim cyan")): 
            a and M.print(*a)
            embed(colors='neutral', stack_depth=1)
            exit_prompt and _Prompt.ask(Text('continue?> ', style="bold magenta"))


class _Prompt(Prompt):
    prompt_suffix = ''
    def get_input(self, console, *a, **kw):
        x = super().get_input(console, *a, **kw)
        console.print('\033[F\033[A', end='')
        return x


def inp(prompt=''):
    """Prompt for input in the terminal. This function is useful for debugging and interactive exploration."""
    with M.pause():
        return _Prompt.ask(Text(prompt or '', style="dim cyan"))


def bp(*a, prompt='breakpoint?> '):
    """Breakpoint"""
    with M.pause():
        a and M.print(*a)
        if not prompt or _Prompt.ask(Text(prompt, style="dim cyan")):
            breakpoint()


def pdb():
    try:
        import pdbr as pdb
    except ImportError:
        try:
            import ipdb as pdb
        except ImportError:
            import pdb
    return pdb


def post_mortem(tb: TracebackType):
    def _print_exc():
        import fnmatch
        rich.console.Console().print_exception(
            suppress=[m for k, m in sys.modules.items() if any(
                fnmatch.fnmatch(k, p) for p in ['fire', 'concurrent.futures', 'threading', 'multiprocessing'])]
        )
        cmds = 'h: help, u: up, d: down, l: code, v: vars, vt: varstree, w: stack, i {var}: inspect'
        rich.print("\n[bold dim]Commands - [/bold dim] " + ", ".join(
            "[bold green]{}[/bold green]:[dim]{}[/dim]".format(*c.split(':')) for c in cmds.split(', ')
        ))

    if tb is not None:
        M.pause(True)
        _print_exc()
        pdb().post_mortem(tb)
        M.pause(False)


def iex(func):
    """Decorator to embed an interactive post-mortem debugger on exception.

    Tries to use `pdbr` if installed; otherwise, prints a rich traceback and re-raises.
    Does not work in subprocesses for obvious reasons.
    """
    import functools
    @functools.wraps(func)
    def outer(*a, **kw):
        _rich_traceback_omit = True
        try:
            return func(*a, **kw)
        except BaseException as e:
            post_mortem(getattr(e, '__traceback__', None))
    return outer


def profile(func=None, **pkw):
    """Decorator to profile a function using pyinstrument."""
    profiler_kw = {k: pkw.pop(k) for k in ['interval', 'use_timing_thread', 'async_mode'] if k in pkw}
    def wrapper(func):
        @functools.wraps(func)
        def outer(*a, **kw):
            import pyinstrument
            p = pyinstrument.Profiler(**(profiler_kw or {}))
            try:
                with p:
                    return func(*a, **kw)
            finally:
                M.pause(True)
                pkw.setdefault('color', True)
                p.print(**pkw)
        return outer
    return wrapper(func) if func is not None else wrapper


def timeit(func=None, **pkw):
    @functools.wraps(func)
    def wrapper(*a, **kw):
        start = time.perf_counter()
        try:
            return func(*a, **kw)
        finally:
            end = time.perf_counter()
            print(f"Function {func.__name__} took {end - start:.4f} seconds")
    return wrapper



''''''
import time
import sys
from functools import wraps
import traceback
from typing import Callable, Iterable, Literal
from concurrent.futures import as_completed
from . import print, T_POOL_MODE, _ttl_pause_wait
from . import utils
from .executor import _get_local, _RemoteTraceback
import mqdm as M

T_DESC_FUNC = Callable[[M.args, int], str]


class mqdm:
    """Create a progress bar for an iterable. (tqdm-like interface)
        
        Args:
            iter (Iterable): The iterable to iterate over.
            desc (str): The description of the progress bar.
            bytes (bool): Whether to show bytes transferred.
            pbar (rich.progress.Progress): An existing progress bar to use.
            leave (bool): Whether to keep/remove the progress bar after completion.
            disable (bool): Whether to disable the progress bar.
            **kw: Additional keyword arguments to pass to the progress bar.
        """
    task_id = None           # the task ID of the progress bar
    pbar = None              # the progress bar instance
    _desc = None             # the description of the progress bar
    _total = None            # the total number of items to iterate over
    _n = 0                   # the number of items completed
    _iter = None             # the item iterator
    # _items = None            # the items of the iterator (for indexing)
    entered = False          # whether the progress bar has called __enter__()
    started = False          # whether the progress bar has beed started (for lazy start)
    _task_dict = None        # the dumped task
    get_desc = None          # a function to get the description
    disable = False          # whether to disable the progress bar

    def __init__(
            self, 
            it=None, desc=None, *, 

            # mqdm arguments
            pool_mode=None, 
            disable=None, 
            # miniters=None, 
            fast_fps_delta=0.05, 
            task_id=None, 

            # progress bar arguments
            progress_kw=None, 
            auto_refresh: bool = True,
            refresh_per_second: float = 8,
            speed_estimate_period: float = 60.0,
            redirect_stdout: bool = True,
            redirect_stderr: bool = True,
            expand: bool = False,

            # rich task arguments
            init_kw=None, 
            start: bool = True,
            total: float|None = None,
            # leave: bool = True,
            # transient: bool = False,
            completed: int = 0,
            visible: bool = True,
            # refresh: bool = False,
            bytes: bool = False,
            **fields,
        ):
        if isinstance(iter, str) and desc is None:  # infer string as description
            it, desc = None, it

        # if disable is not None:
        #     self.disable = disable
        if disable is None:
            disable = self.disable
        self.disable = disable
        # self.miniters = miniters
        self.fast_advance = _speed_increment(delta=fast_fps_delta, disable=disable)

        self._progress_kw = {
            **(progress_kw or {}),
            "auto_refresh": auto_refresh,
            "refresh_per_second": refresh_per_second,
            "speed_estimate_period": speed_estimate_period,
            "redirect_stdout": redirect_stdout,
            "redirect_stderr": redirect_stderr,
            "expand": expand,
        }

        # initialize progress bar task
        kw = {}
        init_kw = {
            **_get_local('defaults', {}), 
            **(init_kw or {}),
            'description': desc,
            'start': start,
            'total': total,
            # 'leave': leave,
            # 'transient': transient,
            'completed': completed,
            'visible': visible,
            'bytes': bytes,
            **fields,
        }
        init_kw = self._process_args(**init_kw)
        if not self.disable:
            pbar = M._get_pbar(pool_mode=pool_mode, **self._progress_kw)
            M._add_instance(self)
            if task_id is None:
                init_kw.setdefault('total', self._total)
                init_kw.setdefault('description', '')
                # if self.miniters:
                #     self._task_dict = M.pbar.new_task(**init_kw)
                # else:
                task_id = pbar.add_task(**init_kw)
            else:
                kw = {**init_kw, **kw}
                if isinstance(task_id, dict):
                    self._task_dict = task_id
                    task_id = task_id['id']
                else:
                    try:
                        task_dict = pbar.dump_task(task_id) or {}
                    except KeyError:
                        task_dict = {}
                self._n = task_dict.get('completed', 0)
                self._total = task_dict.get('total')
                self._desc = task_dict.get('description')
                start = bool(task_dict.get('start_time', start))

        self.task_id = task_id
        self.fast_advance.task_id = task_id
        self.started = start

        # if we have an iterable, start the progress bar
        self(it, desc=..., **kw)

    # ------------------------------ Dunder methods ------------------------------ #

    def __repr__(self) -> str:
        return f'<mqdm[{self.task_id}]({self._n}/{self._total}, {self._desc or ""!r})>'

    def __getstate__(self):
        state = self.__dict__.copy()
        state['get_desc'] = None  # cannot pickle lambda functions
        # state['_items'] = None  # cannot pickle iterators
        state['_iter'] = None  # cannot pickle iterators
        return state

    # def __getitem__(self, i: int):
    #     """Get an item at index."""
    #     return self._items[i]
    
    # def __bool__(self) -> bool:
    #     return self._total is None or self._total

    def __len__(self) -> int:
        """Get the total number of items."""
        return self._total or 0

    def __iter__(self) -> 'mqdm':
        return self

    def __next__(self):
        return next(self._iter)
    
    # ----------------------------- Lifecycle methods ---------------------------- #
    
    def __enter__(self) -> 'mqdm':
        self.entered = True
        return self.open()

    def __exit__(self, t, e, tb):
        self.entered = False
        self.close()
        pbar = M.pbar
        if isinstance(e, KeyboardInterrupt) and pbar is not None:
            pbar.stop()

    def __del__(self):
        try:
            if sys.meta_path is None:
                return 
            self.close()
        except (ImportError, BrokenPipeError, FileNotFoundError) as e:
            pass

    # ----------------------------- Iteration methods ---------------------------- #

    def _get_iter(self, iter, **kw):
        i = x = ...
        try:
            fast_advance = self.fast_advance
            it = enumerate(iter)
            i, x = next(it)
            _ttl_pause_wait()
            self._n += 1
            fast_advance(0, arg=x, i=i, flush=True)
            yield x
            for i, x in it:
                _ttl_pause_wait()
                self._n += 1
                fast_advance(1, arg=x, i=i)
                yield x
            # self._n += 1
            fast_advance(1, flush=True)
        except StopIteration:
            pass
        finally:
            fast_advance(0, flush=True)

    def __call__(self, iter, desc=None, total=None, **kw) -> 'mqdm':
        """Iterate over an iterable with a progress bar."""
        if isinstance(iter, str) and desc is None:  # infer string as description
            iter, desc = None, iter
        if iter is None:  # no iterable yet
            return self.update(0, total=total, **kw)
        if isinstance(iter, int):  # implicit range
            iter = range(iter)

        total = utils.try_len(iter, self._total) if total is None else total
        self.update(0, total=total, description=desc, **kw)
        def _with_iter():
            if self.entered:  # already called __enter__, don't call __exit__() when leaving
                yield from self._get_iter(iter, **kw)
                return
            with self:  # call __enter__() and __exit__()
                yield from self._get_iter(iter, **kw)
        # self._items = iter
        self._iter = _with_iter()
        return self
    
    # ------------------------------ Internal methods ----------------------------- #

    def _attach(self):
        """Attach the task to the progress bar."""
        if self.disable: return

        pbar = M._get_pbar(start=False, **self._progress_kw)
        M._add_instance(self)
        if self._task_dict is not None:
            pbar.load_task(self._task_dict)
            self._task_dict = None
        self.set(total=self._total)

    def _detach(self, remove=None, soft=False):
        """Detach the task from the progress bar."""
        pbar = M.pbar
        if self.disable or pbar is None: return

        # stop and remove task
        if self._task_dict is None:
            self._task_dict = pbar.pop_task(self.task_id, remove=remove)
        M._remove_instance(self)
        M._clear_pbar(strict=False, soft=soft)

    def _process_args(self, *, append_total=None, arg=..., i: int=None, **kw) -> dict:
        """Process keyword arguments for updates."""
        kw = {k: v for k, v in kw.items() if v is not ...}
        if 'leave' in kw:  # tqdm compatibility
            kw['transient'] = not kw.pop('leave')

        # get progress and total
        if 'total' in kw:  # keep local copy of total
            self._total = kw['total']
        if append_total:
            kw['total'] = self._total = (self._total or 0) + append_total
        if kw.get('advance'):
            kw['advance'] = int(kw['advance'])
            self._n += kw['advance']
        if kw.get('advance') == 0:
            kw.pop('advance')
        if kw.get('completed') is not None:
            self._n = kw['completed'] = int(kw['completed'])

        # get description
        if 'desc' in kw:  # tqdm compatibility
            kw['description'] = kw.pop('desc')
        if 'description' in kw and callable(kw['description']):  # handle dynamic descriptions
            self.get_desc = kw.pop('description')
            self.fast_advance.get_desc = self.get_desc
        if kw.get('description') is None and self.get_desc is not None and arg is not ...:
            kw['description'] = self.get_desc(arg, i)
        if kw.get('description') is not None:
            self._desc = kw['description']
        if 'description' in kw and kw['description'] is None:  # ignore None descriptions
            kw.pop('description')

        return kw

    # --------------------------- tqdm-like attributes --------------------------- #

    @property
    def n(self) -> int:
        """The number of items completed."""
        return self._n
    
    @n.setter
    def n(self, n: int):
        """Set the number of items completed."""
        self.set(completed=n)

    @property
    def total(self) -> int|None:
        """The total number of items to iterate over."""
        return self._total
    
    @total.setter
    def total(self, total: int|None):
        """Set the total number of items to iterate over."""
        self.set(total=total)

    # ------------------------------ public methods ------------------------------ #

    def open(self):
        """Add the task to the progress bar."""
        self.entered = True
        self._attach()
        return self

    def close(self, remove=None):
        """Remove the task from the progress bar."""
        self.entered = False
        self._detach(remove=remove)
        return self

    def print(self, *a, **kw) -> 'mqdm':
        """Print above the progress bar."""
        M.print(*a, **kw)
        return self

    def set_description(self, desc) -> 'mqdm':
        """Set the description of the progress bar."""
        return self.update(0, description=desc)

    def update(self, advance: int=1, **kw) -> 'mqdm':
        """Increment the progress bar."""
        return self.set(advance=advance, **kw)

    def set(self, **kw) -> 'mqdm':
        """Update progress bar fields."""
        kw = self._process_args(**kw)
        if self.disable: return self

        # update progress bar
        if kw:
            M.pbar.update_(self.task_id, **kw)
        return self


class _speed_increment:
    """Increment the progress bar in a very fast loop.
    Saves time by reducing inter-process IO.
    """
    __slots__ = ('n', 'nc', 't', 'disable', 'delta', 'get_desc', 'task_id')
    def __init__(self, delta=0, disable=False):
        self.delta = delta
        self.n = self.t = 0
        self.nc = ...
        self.get_desc = None
        self.task_id = None
        self.disable = disable

    def flush(self):
        return self(0, flush=True)

    def __call__(self, n: int=1, completed: int=..., arg=..., i=..., flush=False):
        if self.disable:
            return
        
        if completed is not ...:
            self.nc = completed
            self.n = n = 0

        delta = self.delta
        if delta:
            # If the time since the last increment is less than some delta, increment a local counter
            # The only time this fails is if the iterations are highly irregular 
            # (e.g. a bunch of 1000fps followed by a 100 second iteration
            #       - could happen with overwrite=False type scenarios)
            t = time.time()
            last_t = self.t
            n0 = self.n
            n += n0
            if t - last_t < delta and not flush:
                self.n = n
                return

        pbar = M.pbar
        if pbar is None:
            return
        
        kw = {}
        if i is not ...:
            get_desc = self.get_desc
            if get_desc is not None:
                desc = get_desc(arg, i)
                if desc is not None:
                    kw['description'] = desc

        if self.nc is not ...:
            pbar.update(self.task_id, completed=self.nc + n, **kw)
        elif n:
            pbar.update(self.task_id, advance=n, **kw)

        if delta:
            # if n0:
            self.n = 0
            self.nc = ...
            self.t = t


def ipool(
        fn: Callable, 
        iter: Iterable, 
        desc: str|T_DESC_FUNC='', 
        bar_kw: dict|None=None, 
        n_workers: int=8, 
        pool_mode: T_POOL_MODE='process', 
        ordered_: bool=False, 
        squeeze_: bool=True,
        on_error: Literal['finish', 'cancel', 'skip']='cancel',
        **kw) -> Iterable:
    """Execute a function in a process pool with a progress bar for each task.

    Args:
        fn (Callable): The function to execute.
        iter (Iterable): The iterable to iterate over.
        desc (str|T_DESC_FUNC, optional): The description of the main progress bar. 
        bar_kw (dict, optional): Additional keyword arguments for the sub progress bars.
        n_workers (int, optional): The number of workers in the pool. Defaults to 8.
        pool_mode (str, optional): The mode of the pool. Can be 'process', 'thread', 'sequential'. Defaults to 'process'.
        ordered_ (bool, optional): Whether to yield the results in order. Defaults to False for ipool, True for pool.
        squeeze_ (bool, optional): Whether to skip the pool and main progress bar if there is only one item in the iterable. Defaults to True.
        **kw: Additional keyword arguments to pass to the function.
    """
    # no workers, just run sequentially
    n = utils.try_len(iter, -1)
    if squeeze_ and n >= 0 and n_workers > n:
        n_workers = n
    if squeeze_ and n_workers in {0, 1}:
        pool_mode = 'sequential'
    if pool_mode == 'sequential':
        ordered_ = True

    # if the iterable is a single item, just run the function
    if squeeze_ and n == 1:
        arg = utils.args.from_item(iter[0], **kw)
        yield arg(fn)
        return

    bar_kw = bar_kw or {}
    remote_exceptions = {}
    try:
        # initialize progress bars and run the tasks in a process/thread pool
        with mqdm(desc=desc, elapsed_speed=True, **bar_kw) as pbars:
            with M.executor(pool_mode, bar_kw={'transient': True, **bar_kw}, max_workers=n_workers) as executor:
                try:
                    futures = []
                    for arg in iter:
                        arg_i = utils.args.from_item(arg)
                        arg = utils.args.from_item(arg, **kw)
                        f = executor.submit(fn, *arg.a, **arg.kw)
                        futures.append(f)
                        f.arg = arg_i
                        pbars.set(append_total=1)

                    # get function results
                    for i, f in enumerate(futures if ordered_ else as_completed(futures)):
                        try:
                            x = f.result()
                            pbars.update(arg=f.arg, i=i)
                            yield x
                        except Exception as e:
                            pbars.update(arg=f.arg, i=i)
                            if on_error == 'skip':
                                _add_func_args_str_to_exception(e, fn, f.arg)
                                M.print(traceback.format_exc())
                            elif on_error == 'cancel':
                                _add_func_args_str_to_exception(e, fn, f.arg)
                                for future in futures:
                                    if not future.done():
                                        future.cancel()
                                executor.shutdown(wait=False, cancel_futures=True)
                                raise
                            else:
                                # XXX: this will swallow return values for anything after the error
                                # M.print(traceback.format_exc())
                                # raise
                                _append_remote_exception(remote_exceptions, e, fn, f.arg)
                except KeyboardInterrupt:
                    M.pause()
                    if pool_mode == 'process':
                        try:
                            for pid, process in getattr(executor, '_processes', {}).items():
                                print(f"Killing process {pid}...")
                                process.kill()
                        except Exception as e:
                            M.print(f"Error killing processes: {e}")
                    executor.shutdown(wait=False, cancel_futures=True)
                    raise

    except:
        # pause the progress bar so it doesn't interfere with the traceback
        M.pause()
        raise

    if remote_exceptions:
        raise _combine_remote_exceptions(remote_exceptions)


def _add_func_args_str_to_exception(e, fn, arg):
    """Add the function name and arguments to the exception."""
    cause = getattr(e, '__cause__', None)
    if cause is not None:
        if getattr(cause, 'tb', None) is not None:
            cause.tb = f"{cause.tb}\n\nError Thrown in Remote Function: {fn.__name__}{arg}"
    return e


def _append_remote_exception(excs, e, fn, arg):
    tb = getattr(getattr(e, '__cause__', None), 'tb', None)
    if tb is not None and isinstance(tb, str):
        k = (e.__class__, str(e), tb)
        if k not in excs:
            excs[k] = []
        excs[k].append(f"{fn.__name__}{arg}")
    return excs

def _combine_remote_exceptions(excs, n_fn_limit=10, n_tb_limit=10):
    """Merge the exception arguments into a single string."""

    # TODO: turn this into a class
    # TODO: use type from returned exception (for isinstance) if there is only one type of exception
    # TODO: store fn and args in the exception somewhere

    merged = []
    n_fns = sum(len(args) for args in excs.values())
    for (t, st, tb), args in sorted(excs.items(), key=lambda x: len(x[0]), reverse=True)[:n_tb_limit]:
        arg_str = '\n'.join(args[:n_fn_limit])
        merged.append(f"{tb}\n Seen in {len(args)} Remote Function(s): \n{arg_str}")
        if len(args) > n_fn_limit:
            merged.append(f"... and {len(args) - n_fn_limit} more.")

    if len(excs) > n_tb_limit:
        merged.append(f"... and {len(excs) - n_tb_limit} more exceptions.")

    if len(excs) == 1:
        (t, st, tb), args = next(iter(excs.items()))
        msg = f"{st} -- observed in {n_fns} remote function calls."
        cls = t
    else:
        msg = f"{len(excs)} distinct exceptions observed in {n_fns} remote function calls."
        cls = RuntimeError

    e = cls(msg)
    e.__cause__ = _RemoteTraceback("\n".join(merged))
    return e


@wraps(ipool, ['__doc__'])
def pool(
        fn: Callable, 
        iter: Iterable, 
        desc: str|T_DESC_FUNC='', 
        bar_kw: dict|None=None, 
        n_workers: int=8, 
        pool_mode: T_POOL_MODE='process', 
        results_: list=None,
        ordered_: bool=True, 
        squeeze_: bool=True,
        **kw) -> Iterable:
    results_ = [] if results_ is None else results_
    for x in ipool(fn, iter, desc=desc, bar_kw=bar_kw, n_workers=n_workers, pool_mode=pool_mode, ordered_=ordered_, squeeze_=squeeze_, **kw):
        results_.append(x)
    return results_
